with Interfaces; use Interfaces;
with Ada.Text_IO; use Ada.Text_IO;
with Ada.Integer_Text_IO;
with Ada.Float_Text_IO;

package body Shardware is

  package U_IO is new Ada.Text_IO.Modular_IO (Unsigned_32);

  function TempBytes (Instruction : ByteArr) return ByteArr is 
    TheBytes : ByteArr (1 .. 4);
  begin
    TheBytes (1) := Instruction (7);
    TheBytes (2) := Instruction (8);
    TheBytes (3) := Instruction (9);
    TheBytes (4) := Instruction (10);    
    return TheBytes;
  end TempBytes;

  function UnToSignedInt (Value : Unsigned_32) return Integer is 
    Diff1 : Integer := 0;
    Diff2 : Integer := 0;
  begin
    if Value > 2147483647 then
      Diff1 := Integer(Value - 2147483648);
      Diff2 := Diff1 - 2147483647;
      return Diff2 - 1;
    else
      return Integer(Value);
    end if;
  end UnToSignedInt;

  function IntToUnSigned (Value : Integer) return Unsigned_32 is
    Result : Integer := 0;
  begin
    if Value < 0 then
      Result := Value + 1073741824;
      Result := Result + 1073741824;
      return Unsigned_32(Result) + 2147483648;
    else
      return Unsigned_32(Value);
    end if;
  end IntToUnSigned;

  function BytesToBits (TheBytes : ByteArr) return BitArr is 
    CurByte : Byte := 0; 
    Buff : BitArr (1 .. 32);
    Counter : Integer := 0;
  begin
    for i in 1 .. 32 loop 
      Buff (i) := False;
    end loop;
  for i in 1 .. 4 loop
      CurByte := TheBytes (Unsigned_32(i));
      for j in reverse 1 .. 8 loop
        if CurByte mod 2 > 0 then
          Buff (j + Counter) := True;
        else
          Buff (j + Counter) := False;
        end if;
        CurByte := CurByte / 2;
      end loop;
      Counter := Counter + 8;
    end loop;
    return Buff;
  end BytesToBits;

  function BitsToBytes (BinNumber : BitArr) return ByteArr is
    ReverseBin : BitArr (1 .. 32);
    I : Integer := 1;
    CurByte : Byte := 0;
    Buff : ByteArr (1 .. 4);
    Counter : Integer := 24;
  begin
    for k in reverse 1 .. 32 loop
      ReverseBin (I) := BinNumber (k); -- TODO: Unbloat this mess somehow?
      I := I + 1;
    end loop;
    for i in 1 .. 4 loop
      CurByte := 0;
      for j in reverse 1 .. 8 loop
        if ReverseBin (j + Counter) then 
          CurByte := CurByte + (2 ** (j - 1));
        end if;
      end loop;
      Counter := Counter - 8;
      Buff (Unsigned_32(i)) := CurByte;
    end loop;
    return Buff;
  end BitsToBytes;

  function MyShiftU (Value : Byte; Amount : Integer) return Unsigned_32 is 
  begin
    return Unsigned_32(Shift_Left(Unsigned_128(Value), Amount));
  end MyShiftU;

  function BytesToU (TheBytes : ByteArr) return Unsigned_32 is
  begin
    return Unsigned_32(MyShiftU(TheBytes (1), 24) 
    + MyShiftU(TheBytes (2), 16)
    + MyShiftU(TheBytes (3), 8) 
    + Unsigned_32(TheBytes (4)));
  end BytesToU;

  function BytesToI (TheBytes : ByteArr) return Integer is
  begin
    return UnToSignedInt(BytesToU(TheBytes));
  end BytesToI;

  function BytesToF (TheBytes : ByteArr) return Float is
    Zero : Float := 0.0;
    Result : Float := 0.0;
    TheBits : BitArr (1 .. 32);
    BitBuff : BitArr (1 .. 32);
    Mantissa : Float := 0.0;
    Exponent : Natural := 0;
    Sign : Float := 1.0;
  begin
    TheBits := BytesToBits(TheBytes);

    if TheBits (1) then
      Sign := -1.0;
    else
      Sign := 1.0;
    end if;

    for i in 1 .. 32 loop 
      if i < 10 then
        null;
      else
        if TheBits (i) then
          Mantissa := Mantissa + 2.0 ** (-1 * (i - 9));
        end if;
      end if;
    end loop;
    
    for i in 1 .. 32 loop 
      if i < 25 then
        BitBuff (i) := False;
      else
        BitBuff (i) := TheBits (i - 23);
      end if;
    end loop;

    Exponent := Natural(BytesToU(BitsToBytes(BitBuff)));

    if Exponent > 0 and Exponent < 254 then
      Mantissa := Mantissa + 1.0;
      Result := Sign * Mantissa * (2.0 ** (Exponent - 127));
    elsif Exponent = 255 and Mantissa = 0.0 then
      Result := Sign * 1.0 / Zero;
    elsif Exponent = 0 then 
      Result := 1.0 + Sign * Mantissa * (2.0 ** (Exponent - 127));
    else 
      Result := 0.0 / Zero;
    end if;

    return Result;
  end BytesToF;

  function UToBytes (Number : Unsigned_32) return ByteArr is 
    BitBuff : BitArr (1 .. 32);
    NumBuff : Unsigned_32 := Number;
  begin
    for i in reverse 1 .. 32 loop
      if NumBuff mod 2 > 0 then
        BitBuff (i) := True;
      else
        BitBuff (i) := False;
      end if;
      NumBuff := NumBuff / 2;
    end loop;
    return BitsToBytes(BitBuff);
  end UToBytes;

  function IToBytes (Number : Integer) return ByteArr is 
  begin
    return UToBytes(IntToUnSigned(Number));
  end IToBytes;

  function FToBytes (Number : Float) return ByteArr is 
    ByteBuff : ByteArr (1 .. 4);
  begin
    raise Constraint_Error with "Not Implemented!";
    return ByteBuff;
  end FToBytes;

  function ReadMem (Memory : ByteArr; Address : Unsigned_32; Length : Unsigned_32) return ByteArr is 
    Result : ByteArr (1 .. Length);
  begin
    for i in 1 .. Length loop 
      Result (i) := Memory (Address - 1 + i);
    end loop;
    return Result;
  end ReadMem;

  procedure WriteMem (Memory : in out ByteArr; Address : Unsigned_32; Stuff : ByteArr; ReverseFlag : Boolean := False) is 
  begin
    if ReverseFlag then
      for i in reverse 1 .. Stuff'Length loop 
        Memory (Address + Unsigned_32(i)) := Stuff (Unsigned_32(i));
      end loop;
    else
      for i in 1 .. Stuff'Length loop 
        Memory (Address + Unsigned_32(i)) := Stuff (Unsigned_32(i));
      end loop;
    end if;
  end WriteMem;

  procedure Trap (Registers : RegArrU; Memory : in out ByteArr; MemorySize : in out Unsigned_32; InputTypeFlag : InputType) is 
  begin
    case Registers(0) is
      when 0 => 
        declare
          Pointer : Unsigned_32 := Registers (4) - 1;
          Length : Integer := Integer(Registers (3));
          CharBuff : String (1 .. Length);
        begin
          if InputTypeFlag = LEString then
            Get_Line(CharBuff, Length);
            for i in 1 .. Length loop
              Memory (Pointer + Unsigned_32(i)) := Character'Pos(CharBuff (i));
            end loop;
            Memory (Pointer + Unsigned_32(Length) + 1) := 10;
          elsif InputTypeFlag = LEUint then
            declare
              ByteBuff : ByteArr (1 .. 4);
              ValU : Unsigned_32 := 0;
            begin
              U_IO.Get(ValU);
              ByteBuff := UToBytes(ValU);
              WriteMem(Memory, Pointer, ByteBuff, True);
            end;
          elsif InputTypeFlag = LEInt then
            declare
              ByteBuff : ByteArr (1 .. 4);
              ValI : Integer := 0;
            begin
              Ada.Integer_Text_IO.Get(ValI);
              ByteBuff := IToBytes(ValI);
              WriteMem(Memory, Pointer, ByteBuff, True);
            end;
          elsif InputTypeFlag = LEFloat then
            declare
              ByteBuff : ByteArr (1 .. 4);
              ValF : Float := 0.0;
            begin
              Ada.Float_Text_IO.Get(ValF);
              ByteBuff := FToBytes(ValF);
              WriteMem(Memory, Pointer, ByteBuff, True);
            end;
          else
            null;
          end if;
        end;
      when 1 => 
        declare
          Pointer : Unsigned_32 := Registers (4) - 1;
          Length : Integer := Integer(Registers (3));
          CharBuff : String (1 .. Length);
        begin
          for i in 1 .. Length loop
            CharBuff (i) := Character'Val(Memory (Pointer + Unsigned_32(i)));
          end loop;
          Put(CharBuff);
        end;
      when others => null;
    end case;
  end Trap;

  procedure MovU (Register : in out Unsigned_32; Value : Unsigned_32) is 
  begin
    Register := Value;
  end MovU;

  procedure MovI (Register : in out Integer; Value : Integer) is
  begin
    Register := Value;
  end MovI;

  procedure MovF (Register : in out Float; Value : Float) is 
  begin
    Register := Value;
  end MovF;

  procedure AddU (Register : in out Unsigned_32; Value : Unsigned_32) is 
  begin
    Register := Register + Value;
  end AddU;

  procedure AddI (Register : in out Integer; Value : Integer) is
  begin
    Register := Register + Value;
  end AddI;

  procedure AddF (Register : in out Float; Value : Float) is 
  begin
    Register := Register + Value;
  end AddF;

  procedure SubU (Register : in out Unsigned_32; Value : Unsigned_32) is 
  begin
    if Value > Register then
      raise Constraint_Error with "Unsigned value can't be lower than 0.";
    else
      Register := Register - Value;
    end if;
  end SubU;

  procedure SubI (Register : in out Integer; Value : Integer) is
  begin
    Register := Register - Value;
  end SubI;

  procedure SubF (Register : in out Float; Value : Float) is 
  begin
    Register := Register - Value;
  end SubF;

  procedure MulU (Register : in out Unsigned_32; Value : Unsigned_32) is 
  begin
    Register := Register * Value;
  end MulU;

  procedure MulI (Register : in out Integer; Value : Integer) is
  begin
    Register := Register * Value;
  end MulI;

  procedure MulF (Register : in out Float; Value : Float) is 
  begin
    Register := Register * Value;
  end MulF;

  procedure DivU (Register : in out Unsigned_32; Value : Unsigned_32) is 
  begin
    if Value = 0 then
      raise Constraint_Error with "Please do not divide by 0.";
    else
      Register := Register / Value;
    end if;
  end DivU;

  procedure DivI (Register : in out Integer; Value : Integer) is
  begin
    if Value = 0 then
      raise Constraint_Error with "Please do not divide by 0.";
    else
      Register := Register / Value;
    end if;
  end DivI;

  procedure DivF (Register : in out Float; Value : Float) is 
  begin
    if Value = 0.0 then
      raise Constraint_Error with "Please do not divide by 0.";
    else
      Register := Register / Value;
    end if;
  end DivF;

  function CmpU (Register : Unsigned_32; Value : Unsigned_32) return Boolean is 
  begin
    if Register = Value then
      return True;
    else
      return False;
    end if;
  end CmpU;

  function CmpI (Register : Integer; Value : Integer) return Boolean is 
  begin
    if Register = Value then
      return True;
    else
      return False;
    end if;
  end CmpI;

  function CmpF (Register : Float; Value : Float) return Boolean is 
  begin
    if Register = Value then
      return True;
    else
      return False;
    end if;
  end CmpF;

  function Jump (Address : Unsigned_32) return Unsigned_32 is 
  begin
    if Address mod 2 = 0 then
      return Address;
    else
      raise Constraint_Error with "Can only jump to start of the line";
    end if;
  end Jump;

end Shardware;
