with Ada.Sequential_IO;
with Interfaces; use Interfaces;

package Shardware is

  type Byte is mod 256;
  package Bin_IO is new Ada.Sequential_IO (Byte);
  type ByteArr is array (Unsigned_32 range <>) of Byte;

  type BitArr is array (Integer range <>) of Boolean;

  type RegArrI is array (Integer range <>) of Integer;
  type RegArrF is array (Integer range <>) of Float;
  type RegArrU is array (Integer range <>) of Unsigned_32;

  type InputType is (LEString, LEUint, LEInt, LEFloat);

  function TempBytes (Instruction : ByteArr) return ByteArr;

  function UnToSignedInt (Value : Unsigned_32) return Integer;
  function IntToUnSigned (Value : Integer) return Unsigned_32;

  function BytesToBits (TheBytes : ByteArr) return BitArr;
  function BitsToBytes (BinNumber : BitArr) return ByteArr;

  function MyShiftU (Value : Byte; Amount : Integer) return Unsigned_32;

  function BytesToU (TheBytes : ByteArr) return Unsigned_32;
  function BytesToI (TheBytes : ByteArr) return Integer;
  function BytesToF (TheBytes : ByteArr) return Float;

  function UToBytes (Number : Unsigned_32) return ByteArr;
  function IToBytes (Number : Integer) return ByteArr;
  function FToBytes (Number : Float) return ByteArr;

  function ReadMem (Memory : ByteArr; Address : Unsigned_32; Length : Unsigned_32) return ByteArr;
  procedure WriteMem (Memory : in out ByteArr; Address : Unsigned_32; Stuff : ByteArr; ReverseFlag : Boolean := False);

  procedure Trap (Registers : RegArrU; Memory : in out ByteArr; MemorySize : in out Unsigned_32; InputTypeFlag : InputType);

  procedure MovU (Register : in out Unsigned_32; Value : Unsigned_32);
  procedure MovI (Register : in out Integer; Value : Integer);
  procedure MovF (Register : in out Float; Value : Float);

  procedure AddU (Register : in out Unsigned_32; Value : Unsigned_32);
  procedure AddI (Register : in out Integer; Value : Integer);
  procedure AddF (Register : in out Float; Value : Float);

  procedure SubU (Register : in out Unsigned_32; Value : Unsigned_32);
  procedure SubI (Register : in out Integer; Value : Integer);
  procedure SubF (Register : in out Float; Value : Float);

  procedure MulU (Register : in out Unsigned_32; Value : Unsigned_32);
  procedure MulI (Register : in out Integer; Value : Integer);
  procedure MulF (Register : in out Float; Value : Float);

  procedure DivU (Register : in out Unsigned_32; Value : Unsigned_32);
  procedure DivI (Register : in out Integer; Value : Integer);
  procedure DivF (Register : in out Float; Value : Float);

  function CmpU (Register : Unsigned_32; Value : Unsigned_32) return Boolean;
  function CmpI (Register : Integer; Value : Integer) return Boolean;
  function CmpF (Register : Float; Value : Float) return Boolean;

  function Jump (Address : Unsigned_32) return Unsigned_32;

end Shardware;
